//=============================================================================
// 쌠 (C) 2003 ATMEL Corporation
//
// ̧:		OutputWriter.cpp
// ߲:		Microsoft Visual C++ 6.0
// o͗:
// 쐬:		4-Feb-2003	JP (ATMEL̨)
// ύX:
//
// xҰ:		avr@atmel.com
//
// :		OutputWriter͈ȉ3̧ق𐶐邱Ƃł܂B
//				- ް۰ނ`Ԑݒ肷'bootldr'pͯ ̧
//				- Mް𐳂ǂ邽߂'bootldr'p̧̌
//				- ŖړIΏAVRɑ邱ƂłDESÍ̧
//
// ̑:
//=============================================================================

#include "OutputWriter.h"
#include "MemoryMap.h"
#include "DES.h"
#include "CreateException.h"
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>

const int signatureSize = 4;
const int recordHeaderSize = 6;
const int cipherBlockSize = 8;
const int endOfFrameRecordSize = 1;

//=============================================================================
// ݽ׸

OutputWriter::OutputWriter(MemoryMap *flash, MemoryMap *eeprom,
	unsigned int pageSize, DataBuffer *key, DataBuffer& initialVector,
	DataBuffer& signature, bool updateLockBits, unsigned char lockBits,
	unsigned int memSize, int keyCount, bool createFlashCRC, bool eraseFlash,
	bool nonsenseRecords
	) :
	pageSizeM(pageSize),
	memSizeM(memSize),

	// őڰїeʂAVR߰ޗeʂvZ邱Ƃł܂B
	frameMaxSizeM(
		((signatureSize + recordHeaderSize + pageSize + endOfFrameRecordSize)
			+ (cipherBlockSize - 1)
		) / cipherBlockSize * cipherBlockSize),

	// 1L^̍őe(1ڰт̍ő{̗)
	recordMaxSizeM(frameMaxSizeM - signatureSize - endOfFrameRecordSize),

	// MɏՕpɖړIΏAVRSRAM犄蓖ĂȂ΂ȂȂ޲Đ
	bufferSizeM(2 + frameMaxSizeM + 2),

	signatureM(DataBuffer(signature)),
	initialVectorM(DataBuffer(initialVector)),

	desM(new Des(initialVector)),
	flashM(flash),
	eepromM(eeprom),

	keyCountM(keyCount),
	keyM(key),

	updateLockBitsM(updateLockBits),
	lockBitsM(lockBits),

	frameBufferM(DataBuffer()),

	createFlashCRCM(createFlashCRC),
	eraseFlashM(eraseFlash),
	nonsenseProbabilityM(nonsenseRecords ? 1 : 0)

{
	desM->scheduleKey(key);

	if (createFlashCRCM && flashM)
	{
		// p̈CRCvZĂׯ ؂MemoryMapɑ}
		unsigned int address;
		unsigned short crcCode = 0;
		DataBuffer crcBuffer;

		for (address = 0; address < memSizeM - 2; address++)
			crcCode = crc(flashM->at(address), crcCode);

		crcCode = crc(0, crcCode);
		crcCode = crc(0, crcCode);

		crcBuffer += (unsigned char)(crcCode >> 8);
		crcBuffer += (unsigned char)(crcCode & 0xff);

		flashM->insert(memSizeM - 2, crcBuffer);
	}
}

//=============================================================================
// ޽׸

OutputWriter::~OutputWriter()
{
	delete desM;
}

//=============================================================================
// 'bootldr'pDEŞُ

// ȉ͕KSɏo͂܂(s)!
const char keyFileComment[] =
{
"//==============================================================================\n"
"// ̧:	%s\n"
"// ߲:	IAR Atmel AVR C/EC++߲\n"
"// o͗:	%d\n"
"// 쐬:	%s\n"
"//\n"
"// :	̧ق'create'°قɂĐA'bootldr'ۼުĂɊ܂܂ׂ\n"
"//			łB͓`ԐݒpĈÍꂽ̧ق̐ǗpDES\n"
"//			\܂݂܂B\n"
"//\n"
"// gp:	KEY1 = %s\n"
"//			KEY2 = %s\n"
"//			KEY3 = %s\n"
"//==============================================================================\n"
"\n"
"\n"
};

void OutputWriter::writeKeyFile(char *filename)
{
	FILE *keyFile;
	int lines = (keyCountM * 16) - 1;
	int i;

	if (keyFile = fopen(filename, "w"))
	{
		char dateString[32];
		_strdate(dateString);

		// ̧ٖ߽苎
		char filenameStripped[64], filenameExtension[32];
		_splitpath(filename, NULL, NULL, filenameStripped, filenameExtension);
		strcat(filenameStripped, filenameExtension);

		// ߗ̈o
		fprintf(keyFile, keyFileComment, filenameStripped, keyCountM * 16 * 6,
			dateString, keyM[0].toHexString(),
			((keyCountM > 1) ? keyM[1].toHexString() : "N/A"),
			((keyCountM > 2) ? keyM[2].toHexString() : "N/A")
		);

		fprintf(keyFile, "%s unsigned char kTable[%d * 6] = \n{\n",
			((memSizeM > 65536) ? "__farflash" : "__flash"), keyCountM * 16);

		// 1ްo
		if (keyCountM > 0)
			for (i = 0; i < 16; i++, lines--)
				fprintf(keyFile, lines ? "\t%s,\n" : "\t%s\n",
					desM->getK(2, 15 - i).toCHexString());

		// 2ްo
		if (keyCountM > 1)
		{
			fprintf(keyFile, "\n");

			for (i = 0; i < 16; i++, lines--)
				fprintf(keyFile, lines ? "\t%s,\n" : "\t%s\n",
					desM->getK(1, i).toCHexString());
		}

		// 3ްo
		if (keyCountM > 2)
		{
			fprintf(keyFile, "\n");

			for (i = 0; i < 16; i++, lines--)
				fprintf(keyFile, lines ? "\t%s,\n" : "\t%s\n",
					desM->getK(0, 15 - i).toCHexString());
		}

		fprintf(keyFile, "};\n");

		fclose(keyFile);
	}
	else
		throw new CreateException(ERROR_CREATE_FILE, filename);
}

//=============================================================================
// AVRDESɑ΂ް۰ނƎ̂ɕKvȑSĂ̏܂'bootl
// dr'pͯ ̧ق

const char headerFileComment[] =
{
"//==============================================================================\n"
"// ̧:	%s\n"
"// ߲:	IAR Atmel AVR C/EC++߲\n"
"// o͗:	-\n"
"// 쐬:	%s\n"
"//\n"
"// :	̧ق'create'°قɂĐA'bootldr'ۼުĂɲݸٰނ\n"
"//			ׂłB͈Í̧قŎgp`Ԑݒɏ]ް۰ނ`\n"
"//			Ԑݒ肷ݒ܂݂܂B\n"
"//==============================================================================\n"
"\n"
"\n"
};

void OutputWriter::writeHeaderFile(char *filename)
{
	FILE *headerFile;

	if (headerFile = fopen(filename, "w"))
	{
		char dateString[32];
		_strdate(dateString);

		// ̧ٖ߽苎
		char filenameStripped[64], filenameExtension[32];
		_splitpath(filename, NULL, NULL, filenameStripped, filenameExtension);
		strcat(filenameStripped, filenameExtension);

		// ̧ق̐擪ɒ߂
		fprintf(headerFile, headerFileComment, filenameStripped, dateString);

		// 'filename.ext''FILENAME_EXT'ɕϊ
		_strupr(filenameStripped);
		char *dotIndex = strchr(filenameStripped, '.');
		*dotIndex = '_';

		fprintf(headerFile, "#ifndef %s\n", filenameStripped);
		fprintf(headerFile, "#define %s\n\n\n", filenameStripped);

		// #define`
		fprintf(headerFile, "#define PAGE_SIZE %d\n", pageSizeM);
		fprintf(headerFile, "#define MEM_SIZE %d\n", memSizeM);
		
		if (createFlashCRCM)
			fprintf(headerFile, "#define CRC_CHECK\n");

		fprintf(headerFile, "#define SIGNATURE 0x%s\n",
			signatureM.toHexString());

		fprintf(headerFile, "#define BUFFER_SIZE %d\n", bufferSizeM);

		fprintf(headerFile, "#define INITIALVECTOR_HI 0x%s\n",
			initialVectorM.toHexString(0, 4));

		fprintf(headerFile, "#define INITIALVECTOR_LO 0x%s\n",
			initialVectorM.toHexString(4, 4));

		fprintf(headerFile, "#define KEY_COUNT %d\n", keyCountM);

		fprintf(headerFile, "\n\n#endif // %s\n", filenameStripped);

		fclose(headerFile);
	}
}

//=============================================================================
// 1ڰт̏Ə

void OutputWriter::writeFrame()
{
	// ްȂȂA݂Ȃ
	if (frameBufferM.size() == 0)
		return;

	// ڰт̑OʂɎ[}
	frameBufferM.insert(0, signatureM);

	// 'ڰяI'ڰё}
	frameBufferM += (unsigned char)0x00;

	if (keyCountM > 0)
		desM->encryptBuffer(frameBufferM);

	// CRCvZ
	unsigned int crcCode = 0;

	for (unsigned char *i = frameBufferM.begin(); i != frameBufferM.end(); i++)
		crcCode = crc(*i, crcCode);
	
	crcCode = crc(0, crcCode);
	crcCode = crc(0, crcCode);

	// CRCǉ
	frameBufferM += (unsigned char)(crcCode >> 8);
	frameBufferM += (unsigned char)(crcCode & 0xff);

	// ڰїeʁǍ(CRC܂)ڰт
	fprintf(outputFileM, "%04X", frameBufferM.size());
	fprintf(outputFileM, "%s", frameBufferM.toHexString());

	// ڰїpڰъɏՕ
	frameBufferM.erase();
}

//=============================================================================
// ڰѓɋL^}BڰтɖtȂAڰтĐVڰтJnB

void OutputWriter::insertRecord(DataBuffer& record)
{
	// ۂ̋L^̑Oɂ̕smL^쐬邩ǂ?
	if ((float)rand()/RAND_MAX < nonsenseProbabilityM)
	{
		// smL^搔pwz֐gp
		// (ƁA菭ȂsmL^搔D)
		for (
			float probability = 1.0;
			(float)rand()/RAND_MAX < probability;
			probability = 0.93 * probability)
		{
			// smL^p̋󂫂Ȃ΁Aڰя
			if (frameBufferM.size() == recordMaxSizeM)
				writeFrame();

			// sm1޲Ăǉ
			frameBufferM += 
				(unsigned char)(
					rand() % (TYPE_NONSENSE_LAST - TYPE_NONSENSE_FIRST + 1) +
						TYPE_NONSENSE_FIRST
				);
		}
	}

	// smL^쐬AƎۂ̊eL^̕smL^̂ႢύX
	// ̌Ɏwz֐̊mgpÂ(̧ق̎ceՂɐ
	// Ƃł)o̧ق̐擪ł̕smL^쐬̂荂m܂B
	nonsenseProbabilityM = nonsenseProbabilityM * 0.9;

	if (frameBufferM.size() + record.size() > recordMaxSizeM)
		writeFrame();

	frameBufferM.append(record);

	delete &record;
}

//=============================================================================
// p̈XVް

void OutputWriter::writeFlash()
{
	// ߰ ڽ
	unsigned int pageAddress;

	// ް܂ލŏ߰ނ̊Jnڽ
	unsigned int lowPage = flashM->getLowestAddress() / pageSizeM * pageSizeM;

	// ް܂ލŌ߰ނ̌߰ނ̊Jnڽ
	unsigned int highPage =
		flashM->getHighestAddress() / pageSizeM * pageSizeM + pageSizeM - 1;

	for (pageAddress = lowPage; pageAddress <= highPage;
		pageAddress += pageSizeM)
	{
		bool dataOnPage = false;	// ݂߰ނłް̗L?

		unsigned int index = 0;		// ߰ނł̎wWv

		while (index < pageSizeM)
		{
			if (flashM->in(pageAddress + index))
			{
				// ߰ނł̘Aް ޲Đ
				unsigned int lastindex = index;

				while ((lastindex < pageSizeM) &&
					flashM->in(pageAddress + lastindex)
				)
					lastindex++;

				unsigned int count = lastindex - index;

				// ߰ނōŏ̑傫ް򂪌܂?
				//	͂
				//		̉߰ޑŜ𖞂܂?
				//			ˏ̕KvȂ
				//		ׯ ؏Kv?
				//			CSEG_ERASEL^}
				//		Ȃ:
				//			CSEG_PREPAREL^}
				if (!dataOnPage)
				{
					if (count < pageSizeM)
					{
						if (eraseFlashM)
							insertRecord(buildRecordErase(pageAddress));
						else
							insertRecord(buildRecordPrepare(pageAddress));
					}

					dataOnPage = true;
				}

				// L^ɂڰтɑ}
				insertRecord(buildRecordWrite(pageAddress, index, count));

				// wWXV
				index = lastindex;
			}
			else
				index++;
		}

		if (dataOnPage)
			// ߰ންꍇAPROGRAM()L^}
			insertRecord(buildRecordProgram(pageAddress));
		else if (eraseFlashM)
			// ߰ންȂǁAS߰ނ(ERASEL^摗oɂ)
			// ܂B
			insertRecord(buildRecordErase(pageAddress));
	}
}

//=============================================================================
// EEPROMXVް

void OutputWriter::writeEEPROM()
{
	// ްɊւEEPROM؋ԑŜ𑖍
	unsigned int address = eepromM->getLowestAddress();
	
	while (address <= eepromM->getHighestAddress())
	{
		// 'address'EEPROMްL?
		if (eepromM->in(address))
		{
			// Aް ޲Đ
			int lastAddress = address;

			// (܂AA޲Đׯ ߰ޗeʂł1ڰѓɍv
			// 菬𒲂ׂ܂B)
			while (eepromM->in(lastAddress) &&
				((lastAddress - address) < pageSizeM))
				lastAddress++;

			// L^ɂڰтɑ}
			insertRecord(buildRecordEEPROM(address, lastAddress - address));

			address = lastAddress;
		}
		else
			address++;
	}
}

//=============================================================================
// {ޯċL^揑

void OutputWriter::writeLockBits()
{
	insertRecord(buildRecordLockBitsWrite());
}

//=============================================================================
// 'ؾ ĳ'L^揑

void OutputWriter::writeResetSW()
{
	insertRecord(buildRecordResetSW());
}

//=============================================================================
// 'filename'ƖtꂽVÍ̧ق쐬AړIΏAVRōXV
// ׂSް̧ٓɑ}ADES/3DESÍZ@pĂÍ
// ܂B

void OutputWriter::writeOutputFile(char *filename)
{
	if (outputFileM = fopen(filename, "w"))
	{
		// ׯ ؓe
		if (flashM)
			writeFlash();

		// EEPROMe
		if (eepromM)
			writeEEPROM();

		// {ޯď
		if (updateLockBitsM)
			writeLockBits();

		// AVRؾĂ邱Ƃɂ۸ݸނI
		writeResetSW();
		
		// ڰъɏՕo
		writeFrame();

		fclose(outputFileM);
	}
	else 
		throw new CreateException(ERROR_CREATE_FILE, filename);
}

//=============================================================================
// 'ch'CRCv('oldCRC')XV

unsigned short OutputWriter::crc(unsigned char ch, unsigned short oldCRC)
{
	static const unsigned short crcPoly = 0x8005;

	int n;
	unsigned long m;

	m = ((unsigned long)oldCRC << 8) | ch;
	for (n = 0; n < 8; n++)
		if ((m <<= 1) & 0x1000000)
			m ^= ((unsigned int)crcPoly << 8);

	return (unsigned short)(m >> 8);
}

//=============================================================================
// L^\z

DataBuffer& OutputWriter::buildRecord(RecordType type, int address, int size,
									  const DataBuffer& buffer)
{
	DataBuffer& record = *new DataBuffer(6 + buffer.size());

	record[0] = type;
	record[1] = address >> 16;
	record[2] = address >> 8;
	record[3] = address;
	record[4] = size >> 8;
	record[5] = size;

	for (int i = 0; i < buffer.size(); i++)
		record[i + 6] = buffer[i];

	return record;
}

//=============================================================================
// 㑱傫ް򏑂ݗpɉp̈߰ނď邽߂̋L^
// \z

DataBuffer& OutputWriter::buildRecordErase(unsigned int pageAddress)
{
	return buildRecord(TYPE_ERASE, pageAddress, pageSizeM, DataBuffer());
}

//=============================================================================
// 㑱傫ް򏑂ݗpɉp̈߰ނ邽߂̋L^\z

DataBuffer& OutputWriter::buildRecordPrepare(unsigned int pageAddress)
{
	return buildRecord(TYPE_PREPARE, pageAddress, pageSizeM, DataBuffer());
}

//=============================================================================
// p̈ް̑傫ȉ߂̋L^\z

DataBuffer& OutputWriter::buildRecordWrite(unsigned int pageAddress,
										   unsigned int index, size_t length)
{
	DataBuffer buffer(length);

	for (int i = 0; i < length; i++)
		buffer[i] = flashM->at(pageAddress + index + i);

	return buildRecord(TYPE_DATA, index, length, buffer);
}

//=============================================================================
// ׯ ؂ւ߰ ۸ݸ()pL^\z

DataBuffer& OutputWriter::buildRecordProgram(unsigned int pageAddress)
{
	return buildRecord(TYPE_PROGRAM, pageAddress, pageSizeM / 2, DataBuffer());
}

//=============================================================================
// EEPROMް̑傫ȉ߂̋L^\z

DataBuffer& OutputWriter::buildRecordEEPROM(unsigned int address,
											size_t length)
{
	DataBuffer buffer(length);

	for (int i = 0; i < length; i++)
		buffer[i] = eepromM->at(address + i);

	return buildRecord(TYPE_EEPROM, address, length, buffer);
}

//=============================================================================
// {ޯĂ߂̋L^\z

DataBuffer& OutputWriter::buildRecordLockBitsWrite()
{
	return buildRecord(TYPE_LOCKBITS, (int)lockBitsM << 16, rand(),
		DataBuffer());
}

//=============================================================================
// (Ō̋L^őׂ)ĳؾĂ邽߂̋L^\z

DataBuffer& OutputWriter::buildRecordResetSW()
{
	return buildRecord(TYPE_RESET, rand(), rand(), DataBuffer());
}
